/*QuadPlayer Song Optimizer
Copyright (c) 2005 James Byerly

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/
#ifndef PatternInfo__H
#define PatternInfo__H

#include <vector>
#include <limits>
#include "QPSong.h"

struct PatternInfo
{
	QPNoteCoord Start;//Where in the song the first occurance starts
	size_t Length;//The number of notes in this pattern

	typedef std::vector<QPNoteCoord> OccuranceLocationVec;
	OccuranceLocationVec OccuranceLocation;//Where throughout the song this pattern occurs
	int BytesSaved;//Number of bytes using this pattern as a section would save

	//Comparison operators for std::set test if a pattern is 'better'
	//Undefined results when comparing two PatternInfos from different songs
	
	//Return True if:
	//(Bytes Saved not equal)Left saves more bytes than Right
	//(Occurances not equal)Left has fewer occurances
	//(Length not equal)Left is shorter than right
	//(else)Left occurs earlier than Right
	bool operator >(const PatternInfo &p_OtherPattern) const
	{
		if(BytesSaved != p_OtherPattern.BytesSaved)
		{
			return (BytesSaved > p_OtherPattern.BytesSaved);
		}
		else if(OccuranceLocation.size() != p_OtherPattern.OccuranceLocation.size())
		{
			return (OccuranceLocation.size() < p_OtherPattern.OccuranceLocation.size());
		}
		else if(Length != p_OtherPattern.Length)
		{
			return (Length < p_OtherPattern.Length);
		}
		else
		{
			return (Start.Note < p_OtherPattern.Start.Note);
		}
	}
	//The other operators aren't used, but following are definitions relative to '>' in case they're needed
	/*
	bool operator <(const PatternInfo &p_OtherPattern) const
	{
		return (p_OtherPattern > *this);
	}
	bool operator >=(const PatternInfo &p_OtherPattern) const
	{
		return !(p_OtherPattern > *this);
	}
	bool operator <=(const PatternInfo &p_OtherPattern) const
	{
		return !(*this < p_OtherPattern);
	}
	bool operator ==(const PatternInfo &p_OtherPattern) const
	{
		return !(*this > p_OtherPattern) && !(p_OtherPattern > *this);
	}
	bool operator !=(const PatternInfo &p_OtherPattern) const
	{
		return !(*this == p_OtherPattern);
	}
	*/
	void CalcBytesSaved(const QPSong &p_TheSong)
	{
		const int Occurances = int(OccuranceLocation.size());

		//Each note is 6 bytes, so taking a note out saves 6 bytes for each occurance
		//But, each new section adds a section end (6 bytes) and an entry to the playlist (2 bytes)
		BytesSaved = int(6 * Length * (Occurances - 1))	- int(8 * (Occurances + 1));

		//Keep track of where the previous occurance ended
		//two occurances right after eachother don't make a new section either
		//initalize it to the max possible value - a pattern can never occur there because it'd have to be length 1
		QPNoteCoord PrevOccuranceEnd;
		PrevOccuranceEnd.Section = std::numeric_limits<size_t>::max();
		PrevOccuranceEnd.Note = std::numeric_limits<size_t>::max();

        //But, if the occurance is at the start or end of an existing section, it doesn't add a new section
		typedef PatternInfo::OccuranceLocationVec::const_iterator LocationConstIterator;
		LocationConstIterator StartLocation = OccuranceLocation.begin();
		const LocationConstIterator LastStartLocation = OccuranceLocation.end();
		for(;StartLocation != LastStartLocation; ++StartLocation)
		{
			const size_t CurrentSection = StartLocation->Section;
			const size_t EndLocation = StartLocation->Note + Length;
			if((StartLocation->Note == 0)
				|| (p_TheSong.SectionsNotes[CurrentSection].size() <= EndLocation)
				|| ((PrevOccuranceEnd.Section == CurrentSection)
					&& (PrevOccuranceEnd.Note == EndLocation)))
			{
				//This occurance doesn't make a new section, so subtract the 6 bytes of the section end
				//(it still adds a playlist entry though, so don't subtract the full 8)
				BytesSaved += 6;
			}
			PrevOccuranceEnd.Section = CurrentSection;
			PrevOccuranceEnd.Note = EndLocation;
		}
	}
};
#endif